@@ -19,6 +19,24 @@
  *
  */
 
+#include "awesome.h"
+
+#include "banning.h"
+#include "common/atoms.h"
+#include "common/backtrace.h"
+#include "common/version.h"
+#include "common/xutil.h"
+#include "xkb.h"
+#include "dbus.h"
+#include "event.h"
+#include "ewmh.h"
+#include "globalconf.h"
+#include "objects/client.h"
+#include "objects/screen.h"
+#include "spawn.h"
+#include "systray.h"
+#include "xwindow.h"
+
 #include <getopt.h>
 
 #include <locale.h>
@@ -29,6 +47,8 @@
 
 #include <xcb/bigreq.h>
 #include <xcb/randr.h>
+#include <xcb/xcb_atom.h>
+#include <xcb/xcb_aux.h>
 #include <xcb/xcb_event.h>
 #include <xcb/xinerama.h>
 #include <xcb/xtest.h>
@@ -36,23 +56,6 @@
 
 #include <glib-unix.h>
 
-#include "awesome.h"
-#include "spawn.h"
-#include "objects/client.h"
-#include "xwindow.h"
-#include "ewmh.h"
-#include "dbus.h"
-#include "systray.h"
-#include "event.h"
-#include "property.h"
-#include "screen.h"
-#include "luaa.h"
-#include "common/version.h"
-#include "common/atoms.h"
-#include "common/xcursor.h"
-#include "common/xutil.h"
-#include "common/backtrace.h"
-
 awesome_t globalconf;
 
 /** argv used to run awesome */
@@ -69,8 +72,9 @@ static float main_loop_iteration_limit = 0.1;
 void
 awesome_atexit(bool restart)
 {
-    lua_pushboolean(globalconf.L, restart);
-    signal_object_emit(globalconf.L, &global_signals, "exit", 1);
+    lua_State *L = globalconf_get_lua_State();
+    lua_pushboolean(L, restart);
+    signal_object_emit(L, &global_signals, "exit", 1);
 
     /* Move clients where we want them to be and keep the stacking order intact */
     foreach(c, globalconf.stack)
@@ -94,7 +98,7 @@ awesome_atexit(bool restart)
     systray_cleanup();
 
     /* Close Lua */
-    lua_close(globalconf.L);
+    lua_close(L);
 
     /* X11 is a great protocol. There is a save-set so that reparenting WMs
      * don't kill clients when they shut down. However, when a focused windows
@@ -106,6 +110,8 @@ awesome_atexit(bool restart)
             XCB_NONE, XCB_CURRENT_TIME);
     xcb_aux_sync(globalconf.connection);
 
+    xkb_free();
+
     /* Disconnect *after* closing lua */
     xcb_cursor_context_free(globalconf.cursor_ctx);
     xcb_disconnect(globalconf.connection);
@@ -137,7 +143,7 @@ restore_client_order(xcb_get_property_cookie_t prop_cookie)
                 client_idx++;
             }
 
-    luaA_class_emit_signal(globalconf.L, &client_class, "list", 0);
+    luaA_class_emit_signal(globalconf_get_lua_State(), &client_class, "list", 0);
     p_delete(&reply);
 }
 
@@ -173,6 +179,7 @@ scan(xcb_query_tree_cookie_t tree_c)
     tree_c_len = xcb_query_tree_children_length(tree_r);
     xcb_get_window_attributes_cookie_t attr_wins[tree_c_len];
     xcb_get_property_cookie_t state_wins[tree_c_len];
+    xcb_get_geometry_cookie_t geom_wins[tree_c_len];
 
     for(i = 0; i < tree_c_len; i++)
     {
@@ -180,42 +187,30 @@ scan(xcb_query_tree_cookie_t tree_c)
                                                            wins[i]);
 
         state_wins[i] = xwindow_get_state_unchecked(wins[i]);
+        geom_wins[i] = xcb_get_geometry_unchecked(globalconf.connection, wins[i]);
     }
 
-    xcb_get_geometry_cookie_t *geom_wins[tree_c_len];
-
     for(i = 0; i < tree_c_len; i++)
     {
         attr_r = xcb_get_window_attributes_reply(globalconf.connection,
                                                  attr_wins[i],
                                                  NULL);
+        geom_r = xcb_get_geometry_reply(globalconf.connection, geom_wins[i], NULL);
 
         state = xwindow_get_state_reply(state_wins[i]);
 
-        if(!attr_r || attr_r->override_redirect
+        if(!geom_r || !attr_r || attr_r->override_redirect
            || attr_r->map_state == XCB_MAP_STATE_UNMAPPED
            || state == XCB_ICCCM_WM_STATE_WITHDRAWN)
         {
-            geom_wins[i] = NULL;
             p_delete(&attr_r);
+            p_delete(&geom_r);
             continue;
         }
 
+        client_manage(wins[i], geom_r, attr_r);
+
         p_delete(&attr_r);
-
-        /* Get the geometry of the current window */
-        geom_wins[i] = p_alloca(xcb_get_geometry_cookie_t, 1);
-        *(geom_wins[i]) = xcb_get_geometry_unchecked(globalconf.connection, wins[i]);
-    }
-
-    for(i = 0; i < tree_c_len; i++)
-    {
-        if(!geom_wins[i] || !(geom_r = xcb_get_geometry_reply(globalconf.connection,
-                                                              *(geom_wins[i]), NULL)))
-            continue;
-
-        client_manage(wins[i], geom_r, true);
-
         p_delete(&geom_r);
     }
 
@@ -224,12 +219,98 @@ scan(xcb_query_tree_cookie_t tree_c)
     restore_client_order(prop_cookie);
 }
 
+static void
+acquire_WM_Sn(bool replace)
+{
+    xcb_intern_atom_cookie_t atom_q;
+    xcb_intern_atom_reply_t *atom_r;
+    char *atom_name;
+    xcb_get_selection_owner_reply_t *get_sel_reply;
+    xcb_client_message_event_t ev;
+
+    /* Get the WM_Sn atom */
+    globalconf.selection_owner_window = xcb_generate_id(globalconf.connection);
+    xcb_create_window(globalconf.connection, globalconf.screen->root_depth,
+                      globalconf.selection_owner_window, globalconf.screen->root,
+                      -1, -1, 1, 1, 0,
+                      XCB_COPY_FROM_PARENT, globalconf.screen->root_visual,
+                      0, NULL);
+    xwindow_set_class_instance(globalconf.selection_owner_window);
+    xwindow_set_name_static(globalconf.selection_owner_window,
+            "Awesome WM_Sn selection owner window");
+
+    atom_name = xcb_atom_name_by_screen("WM_S", globalconf.default_screen);
+    if(!atom_name)
+        fatal("error getting WM_Sn atom name");
+
+    atom_q = xcb_intern_atom_unchecked(globalconf.connection, false,
+                                               a_strlen(atom_name), atom_name);
+
+    p_delete(&atom_name);
+
+    atom_r = xcb_intern_atom_reply(globalconf.connection, atom_q, NULL);
+    if(!atom_r)
+        fatal("error getting WM_Sn atom");
+
+    globalconf.selection_atom = atom_r->atom;
+    p_delete(&atom_r);
+
+    /* Is the selection already owned? */
+    get_sel_reply = xcb_get_selection_owner_reply(globalconf.connection,
+            xcb_get_selection_owner(globalconf.connection, globalconf.selection_atom),
+            NULL);
+    if (!get_sel_reply)
+        fatal("GetSelectionOwner for WM_Sn failed");
+    if (!replace && get_sel_reply->owner != XCB_NONE)
+        fatal("another window manager is already running (selection owned; use --replace)");
+
+    /* Acquire the selection */
+    xcb_set_selection_owner(globalconf.connection, globalconf.selection_owner_window,
+                            globalconf.selection_atom, XCB_CURRENT_TIME);
+    if (get_sel_reply->owner != XCB_NONE)
+    {
+        /* Wait for the old owner to go away */
+        xcb_get_geometry_reply_t *geom_reply = NULL;
+        do {
+            p_delete(&geom_reply);
+            geom_reply = xcb_get_geometry_reply(globalconf.connection,
+                    xcb_get_geometry(globalconf.connection, get_sel_reply->owner),
+                    NULL);
+        } while (geom_reply != NULL);
+    }
+    p_delete(&get_sel_reply);
+
+    /* Announce that we are the new owner */
+    p_clear(&ev, 1);
+    ev.response_type = XCB_CLIENT_MESSAGE;
+    ev.window = globalconf.screen->root;
+    ev.format = 32;
+    ev.type = MANAGER;
+    ev.data.data32[0] = XCB_CURRENT_TIME;
+    ev.data.data32[1] = globalconf.selection_atom;
+    ev.data.data32[2] = globalconf.selection_owner_window;
+    ev.data.data32[3] = ev.data.data32[4] = 0;
+
+    xcb_send_event(globalconf.connection, false, globalconf.screen->root, 0xFFFFFF, (char *) &ev);
+}
+
+static xcb_generic_event_t *poll_for_event(void)
+{
+    if (globalconf.pending_event) {
+        xcb_generic_event_t *event = globalconf.pending_event;
+        globalconf.pending_event = NULL;
+        return event;
+    }
+
+    return xcb_poll_for_event(globalconf.connection);
+}
+
 static void
 a_xcb_check(void)
 {
     xcb_generic_event_t *mouse = NULL, *event;
 
-    while((event = xcb_poll_for_event(globalconf.connection)))
+    while((event = poll_for_event()))
     {
         /* We will treat mouse events later.
          * We cannot afford to treat all mouse motion events,
@@ -281,10 +362,24 @@ a_glib_poll(GPollFD *ufds, guint nfsd, gint timeout)
     guint res;
     struct timeval now, length_time;
     float length;
+    lua_State *L = globalconf_get_lua_State();
 
     /* Do all deferred work now */
     awesome_refresh();
 
+    /* Check if the Lua stack is the way it should be */
+    if (lua_gettop(L) != 0) {
+        warn("Something was left on the Lua stack, this is a bug!");
+        luaA_dumpstack(L);
+        lua_settop(L, 0);
+    }
+
+    /* Don't sleep if there is a pending event */
+    assert(globalconf.pending_event == NULL);
+    globalconf.pending_event = xcb_poll_for_event(globalconf.connection);
+    if (globalconf.pending_event != NULL)
+        timeout = 0;
+
     /* Check how long this main loop iteration took */
     gettimeofday(&now, NULL);
     timersub(&now, &last_wakeup, &length_time);
@@ -350,7 +445,10 @@ exit_help(int exit_code)
   -h, --help             show help\n\
   -v, --version          show version\n\
   -c, --config FILE      configuration file to use\n\
-  -k, --check            check configuration file syntax\n");
+      --search DIR       add a directory to the library search path\n\
+  -k, --check            check configuration file syntax\n\
+  -a, --no-argb          disable client transparency support\n\
+  -r, --replace          replace an existing window manager\n");
     exit(exit_code);
 }
 
@@ -363,11 +461,13 @@ int
 main(int argc, char **argv)
 {
     char *confpath = NULL;
+    string_array_t searchpath;
     int xfd, i, opt;
     ssize_t cmdlen = 1;
     xdgHandle xdg;
     bool no_argb = false;
     bool run_test = false;
+    bool replace_wm = false;
     xcb_query_tree_cookie_t tree_c;
     static struct option long_options[] =
     {
@@ -375,7 +475,9 @@ main(int argc, char **argv)
         { "version", 0, NULL, 'v' },
         { "config",  1, NULL, 'c' },
         { "check",   0, NULL, 'k' },
+        { "search",  1, NULL, 's' },
         { "no-argb", 0, NULL, 'a' },
+        { "replace", 0, NULL, 'r' },
         { NULL,      0, NULL, 0 }
     };
 
@@ -387,7 +489,9 @@ main(int argc, char **argv)
     p_clear(&globalconf, 1);
     globalconf.keygrabber = LUA_REFNIL;
     globalconf.mousegrabber = LUA_REFNIL;
+    globalconf.exit_code = EXIT_SUCCESS;
     buffer_init(&globalconf.startup_errors);
+    string_array_init(&searchpath);
 
     /* save argv */
     for(i = 0; i < argc; i++)
@@ -405,14 +509,8 @@ main(int argc, char **argv)
     /* Text won't be printed correctly otherwise */
     setlocale(LC_CTYPE, "");
 
-    /* Get XDG basedir data */
-    xdgInitHandle(&xdg);
-
-    /* init lua */
-    luaA_init(&xdg);
-
     /* check args */
-    while((opt = getopt_long(argc, argv, "vhkc:a",
+    while((opt = getopt_long(argc, argv, "vhkc:ar",
                              long_options, NULL)) != -1)
         switch(opt)
         {
@@ -426,16 +524,31 @@ main(int argc, char **argv)
             run_test = true;
             break;
           case 'c':
-            if(a_strlen(optarg))
-                confpath = a_strdup(optarg);
-            else
-                fatal("-c option requires a file name");
+            if (confpath != NULL)
+                fatal("--config may only be specified once");
+            confpath = a_strdup(optarg);
+            break;
+          case 's':
+            string_array_append(&searchpath, a_strdup(optarg));
             break;
           case 'a':
             no_argb = true;
             break;
+          case 'r':
+            replace_wm = true;
+            break;
+          default:
+            exit_help(EXIT_FAILURE);
+            break;
         }
 
+    /* Get XDG basedir data */
+    xdgInitHandle(&xdg);
+
+    /* init lua */
+    luaA_init(&xdg, &searchpath);
+    string_array_wipe(&searchpath);
+
     if (run_test)
     {
         if(!luaA_parserc(&xdg, confpath, false))
@@ -466,6 +579,9 @@ main(int argc, char **argv)
     /* We have no clue where the input focus is right now */
     globalconf.focus.need_update = true;
 
+    /* set the default preferred icon size */
+    globalconf.preferred_icon_size = 0;
+
     /* X stuff */
     globalconf.connection = xcb_connect(NULL, &globalconf.default_screen);
     if(xcb_connection_has_error(globalconf.connection))
@@ -497,6 +613,17 @@ main(int argc, char **argv)
 
     if (xcb_cursor_context_new(globalconf.connection, globalconf.screen, &globalconf.cursor_ctx) < 0)
         fatal("Failed to initialize xcb-cursor");
+    globalconf.xrmdb = xcb_xrm_database_from_default(globalconf.connection);
+    if (globalconf.xrmdb == NULL)
+        globalconf.xrmdb = xcb_xrm_database_from_string("");
+    if (globalconf.xrmdb == NULL)
+        fatal("Failed to initialize xcb-xrm");
+
+    /* Did we get some usable data from the above X11 setup? */
+    draw_test_cairo_xcb();
+
+    /* Acquire the WM_Sn selection */
+    acquire_WM_Sn(replace_wm);
 
     /* initialize dbus */
     a_dbus_init();
@@ -519,7 +646,7 @@ main(int argc, char **argv)
                                                       globalconf.screen->root,
                                                       XCB_CW_EVENT_MASK, &select_input_val);
         if (xcb_request_check(globalconf.connection, cookie))
-            fatal("another window manager is already running");
+            fatal("another window manager is already running (can't select SubstructureRedirect)");
     }
 
     /* Prefetch the maximum request length */
@@ -528,16 +655,16 @@ main(int argc, char **argv)
     /* check for xtest extension */
     const xcb_query_extension_reply_t *query;
     query = xcb_get_extension_data(globalconf.connection, &xcb_test_id);
-    globalconf.have_xtest = query->present;
+    globalconf.have_xtest = query && query->present;
 
     /* check for shape extension */
     query = xcb_get_extension_data(globalconf.connection, &xcb_shape_id);
-    globalconf.have_shape = query->present;
+    globalconf.have_shape = query && query->present;
+
+    event_init();
 
     /* Allocate the key symbols */
     globalconf.keysyms = xcb_key_symbols_alloc(globalconf.connection);
-    xcb_get_modifier_mapping_cookie_t xmapping_cookie =
-        xcb_get_modifier_mapping_unchecked(globalconf.connection);
 
     /* init atom cache */
     atoms_init(globalconf.connection);
@@ -545,11 +672,6 @@ main(int argc, char **argv)
     /* init screens information */
     screen_scan();
 
-    xutil_lock_mask_get(globalconf.connection, xmapping_cookie,
-                        globalconf.keysyms, &globalconf.numlockmask,
-                        &globalconf.shiftlockmask, &globalconf.capslockmask,
-                        &globalconf.modeswitchmask);
-
     /* do this only for real screen */
     ewmh_init();
     systray_init();
@@ -557,6 +679,9 @@ main(int argc, char **argv)
     /* init spawn (sn) */
     spawn_init();
 
+    /* init xkb */
+    xkb_init();
+
     /* The default GC is just a newly created associated with a window with
      * depth globalconf.default_depth.
      * The window_no_focus is used for "nothing has the input focus". */
@@ -575,6 +700,8 @@ main(int argc, char **argv)
                           1,
                           globalconf.default_cmap
                       });
+    xwindow_set_class_instance(globalconf.focus.window_no_focus);
+    xwindow_set_name_static(globalconf.focus.window_no_focus, "Awesome no input window");
     xcb_map_window(globalconf.connection, globalconf.focus.window_no_focus);
     xcb_create_gc(globalconf.connection, globalconf.gc, globalconf.focus.window_no_focus,
                   XCB_GC_FOREGROUND | XCB_GC_BACKGROUND,
@@ -593,6 +720,9 @@ main(int argc, char **argv)
     xcb_ungrab_server(globalconf.connection);
     xcb_flush(globalconf.connection);
 
+    /* get the current wallpaper, from now on we are informed when it changes */
+    root_update_wallpaper();
+
     /* Parse and run configuration file */
     if (!luaA_parserc(&xdg, confpath, true))
         fatal("couldn't find any rc file");
@@ -604,7 +734,7 @@ main(int argc, char **argv)
     /* scan existing windows */
     scan(tree_c);
 
-    xcb_flush(globalconf.connection);
+    luaA_emit_startup();
 
     /* Setup the main context */
     g_main_context_set_poll_func(g_main_context_default(), &a_glib_poll);
@@ -621,7 +751,7 @@ main(int argc, char **argv)
 
     awesome_atexit(false);
 
-    return EXIT_SUCCESS;
+    return globalconf.exit_code;
 }
 
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
