@@ -20,11 +20,10 @@
  */
 
 #include "banning.h"
-#include "objects/tag.h"
-#include "screen.h"
+#include "globalconf.h"
+#include "objects/client.h"
 
 /** Reban windows following current selected tags.
- * \param screen The screen to arrange.
  */
 void
 banning_need_update(void)
