@@ -19,47 +19,49 @@
  *
  */
 
-#include <xcb/xcb.h>
-#include <xcb/randr.h>
-#include <xcb/shape.h>
-#include <xcb/xcb_atom.h>
-#include <xcb/xcb_icccm.h>
-#include <xcb/xcb_event.h>
-
-#include "awesome.h"
 #include "event.h"
+#include "awesome.h"
 #include "property.h"
 #include "objects/tag.h"
 #include "objects/drawin.h"
 #include "xwindow.h"
 #include "ewmh.h"
 #include "objects/client.h"
-#include "keyresolv.h"
 #include "keygrabber.h"
 #include "mousegrabber.h"
 #include "luaa.h"
 #include "systray.h"
-#include "screen.h"
+#include "xkb.h"
+#include "objects/screen.h"
 #include "common/atoms.h"
 #include "common/xutil.h"
 
+#include <xcb/xcb.h>
+#include <xcb/randr.h>
+#include <xcb/shape.h>
+#include <xcb/xcb_atom.h>
+#include <xcb/xcb_icccm.h>
+#include <xcb/xcb_event.h>
+#include <xcb/xkb.h>
+
 #define DO_EVENT_HOOK_CALLBACK(type, xcbtype, xcbeventprefix, arraytype, match) \
     static void \
     event_##xcbtype##_callback(xcb_##xcbtype##_press_event_t *ev, \
                                arraytype *arr, \
+                               lua_State *L, \
                                int oud, \
                                int nargs, \
                                void *data) \
     { \
-        int abs_oud = oud < 0 ? ((lua_gettop(globalconf.L) + 1) + oud) : oud; \
+        int abs_oud = oud < 0 ? ((lua_gettop(L) + 1) + oud) : oud; \
         int item_matching = 0; \
         foreach(item, *arr) \
             if(match(ev, *item, data)) \
             { \
                 if(oud) \
-                    luaA_object_push_item(globalconf.L, abs_oud, *item); \
+                    luaA_object_push_item(L, abs_oud, *item); \
                 else \
-                    luaA_object_push(globalconf.L, *item); \
+                    luaA_object_push(L, *item); \
                 item_matching++; \
             } \
         for(; item_matching > 0; item_matching--) \
@@ -68,18 +70,18 @@
             { \
               case xcbeventprefix##_PRESS: \
                 for(int i = 0; i < nargs; i++) \
-                    lua_pushvalue(globalconf.L, - nargs - item_matching); \
-                luaA_object_emit_signal(globalconf.L, - nargs - 1, "press", nargs); \
+                    lua_pushvalue(L, - nargs - item_matching); \
+                luaA_object_emit_signal(L, - nargs - 1, "press", nargs); \
                 break; \
               case xcbeventprefix##_RELEASE: \
                 for(int i = 0; i < nargs; i++) \
-                    lua_pushvalue(globalconf.L, - nargs - item_matching); \
-                luaA_object_emit_signal(globalconf.L, - nargs - 1, "release", nargs); \
+                    lua_pushvalue(L, - nargs - item_matching); \
+                luaA_object_emit_signal(L, - nargs - 1, "release", nargs); \
                 break; \
             } \
-            lua_pop(globalconf.L, 1); \
+            lua_pop(L, 1); \
         } \
-        lua_pop(globalconf.L, nargs); \
+        lua_pop(L, nargs); \
     }
 
 static bool
@@ -113,16 +115,18 @@ event_handle_mousegrabber(int x, int y, uint16_t mask)
 {
     if(globalconf.mousegrabber != LUA_REFNIL)
     {
-        lua_rawgeti(globalconf.L, LUA_REGISTRYINDEX, globalconf.mousegrabber);
-        mousegrabber_handleevent(globalconf.L, x, y, mask);
-        if(lua_pcall(globalconf.L, 1, 1, 0))
+        lua_State *L = globalconf_get_lua_State();
+        mousegrabber_handleevent(L, x, y, mask);
+        lua_rawgeti(L, LUA_REGISTRYINDEX, globalconf.mousegrabber);
+        if(!luaA_dofunction(L, 1, 1))
         {
-            warn("error running function: %s", lua_tostring(globalconf.L, -1));
-            luaA_mousegrabber_stop(globalconf.L);
+            warn("Stopping mousegrabber.");
+            luaA_mousegrabber_stop(L);
+        } else {
+            if(!lua_isboolean(L, -1) || !lua_toboolean(L, -1))
+                luaA_mousegrabber_stop(L);
+            lua_pop(L, 1);  /* pop returned value */
         }
-        else if(!lua_isboolean(globalconf.L, -1) || !lua_toboolean(globalconf.L, -1))
-            luaA_mousegrabber_stop(globalconf.L);
-        lua_pop(globalconf.L, 1);  /* pop returned value */
         return true;
     }
     return false;
@@ -130,10 +134,11 @@ event_handle_mousegrabber(int x, int y, uint16_t mask)
 
 /** Emit a button signal.
  * The top of the lua stack has to be the object on which to emit the event.
+ * \param L The Lua VM state.
  * \param ev The event to handle.
  */
 static void
-event_emit_button(xcb_button_press_event_t *ev)
+event_emit_button(lua_State *L, xcb_button_press_event_t *ev)
 {
     const char *name;
     switch(XCB_EVENT_RESPONSE_TYPE(ev))
@@ -149,12 +154,12 @@ event_emit_button(xcb_button_press_event_t *ev)
     }
 
     /* Push the event's info */
-    lua_pushnumber(globalconf.L, ev->event_x);
-    lua_pushnumber(globalconf.L, ev->event_y);
-    lua_pushnumber(globalconf.L, ev->detail);
-    luaA_pushmodifiers(globalconf.L, ev->state);
+    lua_pushinteger(L, ev->event_x);
+    lua_pushinteger(L, ev->event_y);
+    lua_pushinteger(L, ev->detail);
+    luaA_pushmodifiers(L, ev->state);
     /* And emit the signal */
-    luaA_object_emit_signal(globalconf.L, -5, name, 4);
+    luaA_object_emit_signal(L, -5, name, 4);
 }
 
 /** The button press event handler.
@@ -163,13 +168,24 @@ event_emit_button(xcb_button_press_event_t *ev)
 static void
 event_handle_button(xcb_button_press_event_t *ev)
 {
+    lua_State *L = globalconf_get_lua_State();
     client_t *c;
     drawin_t *drawin;
 
     globalconf.timestamp = ev->time;
 
-    if(event_handle_mousegrabber(ev->root_x, ev->root_y, 1 << (ev->detail - 1 + 8)))
-        return;
+    {
+        /* ev->state contains the state before the event. Compute the state
+         * after the event for the mousegrabber.
+         */
+        uint16_t state = ev->state, change = 1 << (ev->detail - 1 + 8);
+        if (XCB_EVENT_RESPONSE_TYPE(ev) == XCB_BUTTON_PRESS)
+            state |= change;
+        else
+            state &= ~change;
+        if(event_handle_mousegrabber(ev->root_x, ev->root_y, state))
+            return;
+    }
 
     /* ev->state is
      * button status (8 bits) + modifiers status (8 bits)
@@ -189,13 +205,13 @@ event_handle_button(xcb_button_press_event_t *ev)
         }
 
         /* Push the drawable */
-        luaA_object_push(globalconf.L, drawin);
-        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
+        luaA_object_push(L, drawin);
+        luaA_object_push_item(L, -1, drawin->drawable);
         /* and handle the button raw button event */
-        event_emit_button(ev);
-        lua_pop(globalconf.L, 1);
+        event_emit_button(L, ev);
+        lua_pop(L, 1);
         /* check if any button object matches */
-        event_button_callback(ev, &drawin->buttons, -1, 1, NULL);
+        event_button_callback(ev, &drawin->buttons, L, -1, 1, NULL);
         /* Either we are receiving this due to ButtonPress/Release on the root
          * window or because we grabbed the button on the window. In the later
          * case we have to call AllowEvents.
@@ -207,26 +223,44 @@ event_handle_button(xcb_button_press_event_t *ev)
                              XCB_ALLOW_ASYNC_POINTER,
                              XCB_CURRENT_TIME);
     }
-    else if((c = client_getbyframewin(ev->event)))
+    else if((c = client_getbyframewin(ev->event)) || (c = client_getbywin(ev->event)))
     {
-        luaA_object_push(globalconf.L, c);
-        /* And handle the button raw button event */
-        event_emit_button(ev);
-        /* then check if a titlebar was "hit" */
-        int x = ev->event_x, y = ev->event_y;
-        drawable_t *d = client_get_drawable_offset(c, &x, &y);
-        if (d)
+        /* For clicks inside of c->window, we get two events. Once because of a
+         * passive grab on c->window and then again for c->frame_window.
+         * Ignore the second event (identifiable by ev->child != XCB_NONE).
+         */
+        if (ev->event != c->frame_window || ev->child == XCB_NONE)
         {
-            /* Copy the event so that we can fake x/y */
-            xcb_button_press_event_t event = *ev;
-            event.event_x = x;
-            event.event_y = y;
-            luaA_object_push_item(globalconf.L, -1, d);
-            event_emit_button(&event);
-            lua_pop(globalconf.L, 1);
+            luaA_object_push(L, c);
+            if (c->window == ev->event)
+            {
+                /* Button event into the client itself (not titlebar), translate
+                 * into the frame window.
+                 */
+                ev->event_x += c->titlebar[CLIENT_TITLEBAR_LEFT].size;
+                ev->event_y += c->titlebar[CLIENT_TITLEBAR_TOP].size;
+            }
+            /* And handle the button raw button event */
+            event_emit_button(L, ev);
+            /* then check if a titlebar was "hit" */
+            if (c->frame_window == ev->event)
+            {
+                int x = ev->event_x, y = ev->event_y;
+                drawable_t *d = client_get_drawable_offset(c, &x, &y);
+                if (d)
+                {
+                    /* Copy the event so that we can fake x/y */
+                    xcb_button_press_event_t event = *ev;
+                    event.event_x = x;
+                    event.event_y = y;
+                    luaA_object_push_item(L, -1, d);
+                    event_emit_button(L, &event);
+                    lua_pop(L, 1);
+                }
+            }
+            /* then check if any button objects match */
+            event_button_callback(ev, &c->buttons, L, -1, 1, NULL);
         }
-        /* then check if any button objects match */
-        event_button_callback(ev, &c->buttons, -1, 1, NULL);
         xcb_allow_events(globalconf.connection,
                          XCB_ALLOW_REPLAY_POINTER,
                          XCB_CURRENT_TIME);
@@ -234,7 +268,7 @@ event_handle_button(xcb_button_press_event_t *ev)
     else if(ev->child == XCB_NONE)
         if(globalconf.screen->root == ev->event)
         {
-            event_button_callback(ev, &globalconf.buttons, 0, 0, NULL);
+            event_button_callback(ev, &globalconf.buttons, L, 0, 0, NULL);
             return;
         }
 }
@@ -339,13 +373,15 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
         }
         if(ev->value_mask & XCB_CONFIG_WINDOW_BORDER_WIDTH)
         {
+            lua_State *L = globalconf_get_lua_State();
+
             diff_border = ev->border_width - bw;
             diff_h += diff_border;
             diff_w += diff_border;
 
-            luaA_object_push(globalconf.L, c);
-            window_set_border_width(globalconf.L, -1, ev->border_width);
-            lua_pop(globalconf.L, 1);
+            luaA_object_push(L, c);
+            window_set_border_width(L, -1, ev->border_width);
+            lua_pop(L, 1);
         }
 
         /* If the client resizes without moving itself, apply window gravity */
@@ -363,6 +399,36 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
             /* ICCCM 4.1.5 / 4.2.3, if nothing was changed, send an event saying so */
             client_send_configure(c);
     }
+    else if (xembed_getbywin(&globalconf.embedded, ev->window))
+    {
+        /* Ignore this so that systray icons cannot resize themselves.
+         * We decide their size!
+         * However, Xembed says that we act like a WM to the embedded window and
+         * thus we have to send a synthetic configure notify informing the
+         * window that its configure request was denied.
+         */
+        xcb_get_geometry_cookie_t geom_cookie =
+            xcb_get_geometry_unchecked(globalconf.connection, ev->window);
+        xcb_translate_coordinates_cookie_t coords_cookie =
+            xcb_translate_coordinates_unchecked(globalconf.connection,
+                    ev->window, globalconf.screen->root, 0, 0);
+        xcb_get_geometry_reply_t *geom =
+            xcb_get_geometry_reply(globalconf.connection, geom_cookie, NULL);
+        xcb_translate_coordinates_reply_t *coords =
+            xcb_translate_coordinates_reply(globalconf.connection, coords_cookie, NULL);
+
+        if (geom && coords)
+        {
+            xwindow_configure(ev->window,
+                    (area_t) { .x = coords->dst_x,
+                               .y = coords->dst_y,
+                               .width = geom->width,
+                               .height = geom->height },
+                    0);
+        }
+        p_delete(&geom);
+        p_delete(&coords);
+    }
     else
         event_handle_configurerequest_configure_window(ev);
 }
@@ -373,13 +439,16 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
 static void
 event_handle_configurenotify(xcb_configure_notify_event_t *ev)
 {
-    const xcb_screen_t *screen = globalconf.screen;
+    xcb_screen_t *screen = globalconf.screen;
 
-    if(ev->window == screen->root
-       && (ev->width != screen->width_in_pixels
-           || ev->height != screen->height_in_pixels))
-        /* it's not that we panic, but restart */
-        awesome_restart();
+    if(ev->window == screen->root)
+        globalconf.screen_need_refresh = true;
+
+    /* Copy what XRRUpdateConfiguration() would do: Update the configuration */
+    if(ev->window == screen->root) {
+        screen->width_in_pixels = ev->width;
+        screen->height_in_pixels = ev->height;
+    }
 }
 
 /** The destroy notify event handler.
@@ -403,7 +472,7 @@ event_handle_destroynotify(xcb_destroy_notify_event_t *ev)
 
 /** Record that the given drawable contains the pointer.
  */
-static void
+void
 event_drawable_under_mouse(lua_State *L, int ud)
 {
     void *d;
@@ -445,6 +514,7 @@ event_drawable_under_mouse(lua_State *L, int ud)
 static void
 event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 {
+    lua_State *L = globalconf_get_lua_State();
     drawin_t *w;
     client_t *c;
 
@@ -455,35 +525,35 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 
     if((c = client_getbyframewin(ev->event)))
     {
-        luaA_object_push(globalconf.L, c);
-        lua_pushnumber(globalconf.L, ev->event_x);
-        lua_pushnumber(globalconf.L, ev->event_y);
-        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
+        luaA_object_push(L, c);
+        lua_pushinteger(L, ev->event_x);
+        lua_pushinteger(L, ev->event_y);
+        luaA_object_emit_signal(L, -3, "mouse::move", 2);
 
         /* now check if a titlebar was "hit" */
         int x = ev->event_x, y = ev->event_y;
         drawable_t *d = client_get_drawable_offset(c, &x, &y);
         if (d)
         {
-            luaA_object_push_item(globalconf.L, -1, d);
-            event_drawable_under_mouse(globalconf.L, -1);
-            lua_pushnumber(globalconf.L, x);
-            lua_pushnumber(globalconf.L, y);
-            luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
-            lua_pop(globalconf.L, 1);
+            luaA_object_push_item(L, -1, d);
+            event_drawable_under_mouse(L, -1);
+            lua_pushinteger(L, x);
+            lua_pushinteger(L, y);
+            luaA_object_emit_signal(L, -3, "mouse::move", 2);
+            lua_pop(L, 1);
         }
-        lua_pop(globalconf.L, 1);
+        lua_pop(L, 1);
     }
 
     if((w = drawin_getbywin(ev->event)))
     {
-        luaA_object_push(globalconf.L, w);
-        luaA_object_push_item(globalconf.L, -1, w->drawable);
-        event_drawable_under_mouse(globalconf.L, -1);
-        lua_pushnumber(globalconf.L, ev->event_x);
-        lua_pushnumber(globalconf.L, ev->event_y);
-        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
-        lua_pop(globalconf.L, 2);
+        luaA_object_push(L, w);
+        luaA_object_push_item(L, -1, w->drawable);
+        event_drawable_under_mouse(L, -1);
+        lua_pushinteger(L, ev->event_x);
+        lua_pushinteger(L, ev->event_y);
+        luaA_object_emit_signal(L, -3, "mouse::move", 2);
+        lua_pop(L, 2);
     }
 }
 
@@ -493,6 +563,7 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 static void
 event_handle_leavenotify(xcb_leave_notify_event_t *ev)
 {
+    lua_State *L = globalconf_get_lua_State();
     client_t *c;
 
     globalconf.timestamp = ev->time;
@@ -505,13 +576,14 @@ event_handle_leavenotify(xcb_leave_notify_event_t *ev)
      */
     if(ev->detail != XCB_NOTIFY_DETAIL_INFERIOR && (c = client_getbyframewin(ev->event)))
     {
-        luaA_object_push(globalconf.L, c);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
+        luaA_object_push(L, c);
+        luaA_object_emit_signal(L, -1, "mouse::leave", 0);
+        lua_pop(L, 1);
     }
 
-    lua_pushnil(globalconf.L);
-    event_drawable_under_mouse(globalconf.L, -1);
-    lua_pop(globalconf.L, 1);
+    lua_pushnil(L);
+    event_drawable_under_mouse(L, -1);
+    lua_pop(L, 1);
 }
 
 /** The enter notify event handler.
@@ -520,6 +592,7 @@ event_handle_leavenotify(xcb_leave_notify_event_t *ev)
 static void
 event_handle_enternotify(xcb_enter_notify_event_t *ev)
 {
+    lua_State *L = globalconf_get_lua_State();
     client_t *c;
     drawin_t *drawin;
 
@@ -530,29 +603,29 @@ event_handle_enternotify(xcb_enter_notify_event_t *ev)
 
     if((drawin = drawin_getbywin(ev->event)))
     {
-        luaA_object_push(globalconf.L, drawin);
-        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
-        event_drawable_under_mouse(globalconf.L, -1);
-        lua_pop(globalconf.L, 2);
+        luaA_object_push(L, drawin);
+        luaA_object_push_item(L, -1, drawin->drawable);
+        event_drawable_under_mouse(L, -1);
+        lua_pop(L, 2);
     }
 
     if((c = client_getbyframewin(ev->event)))
     {
-        luaA_object_push(globalconf.L, c);
+        luaA_object_push(L, c);
         /* Ignore enter with detail inferior: The pointer was previously inside
          * of a child window, so technically this isn't a 'real' enter.
          */
         if (ev->detail != XCB_NOTIFY_DETAIL_INFERIOR)
-            luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+            luaA_object_emit_signal(L, -1, "mouse::enter", 0);
 
         drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
         if (d)
         {
-            luaA_object_push_item(globalconf.L, -1, d);
-            event_drawable_under_mouse(globalconf.L, -1);
-            lua_pop(globalconf.L, 1);
+            luaA_object_push_item(L, -1, d);
+            event_drawable_under_mouse(L, -1);
+            lua_pop(L, 1);
         }
-        lua_pop(globalconf.L, 1);
+        lua_pop(L, 1);
     }
     else if (ev->event == globalconf.screen->root) {
         /* When there are multiple X screens with awesome running separate
@@ -624,33 +697,34 @@ event_handle_expose(xcb_expose_event_t *ev)
 static void
 event_handle_key(xcb_key_press_event_t *ev)
 {
+    lua_State *L = globalconf_get_lua_State();
     globalconf.timestamp = ev->time;
 
     if(globalconf.keygrabber != LUA_REFNIL)
     {
-        lua_rawgeti(globalconf.L, LUA_REGISTRYINDEX, globalconf.keygrabber);
-        if(keygrabber_handlekpress(globalconf.L, ev))
+        if(keygrabber_handlekpress(L, ev))
         {
-            if(lua_pcall(globalconf.L, 3, 1, 0))
+            lua_rawgeti(L, LUA_REGISTRYINDEX, globalconf.keygrabber);
+
+            if(!luaA_dofunction(L, 3, 0))
             {
-                warn("error running function: %s", lua_tostring(globalconf.L, -1));
-                luaA_keygrabber_stop(globalconf.L);
+                warn("Stopping keygrabber.");
+                luaA_keygrabber_stop(L);
             }
         }
-        lua_pop(globalconf.L, 1);  /* pop returned value or function if not called */
     }
     else
     {
         /* get keysym ignoring all modifiers */
-        xcb_keysym_t keysym = keyresolv_get_keysym(ev->detail, 0);
+        xcb_keysym_t keysym = xcb_key_symbols_get_keysym(globalconf.keysyms, ev->detail, 0);
         client_t *c;
-        if((c = client_getbywin(ev->event)))
+        if((c = client_getbywin(ev->event)) || (c = client_getbynofocuswin(ev->event)))
         {
-            luaA_object_push(globalconf.L, c);
-            event_key_callback(ev, &c->keys, -1, 1, &keysym);
+            luaA_object_push(L, c);
+            event_key_callback(ev, &c->keys, L, -1, 1, &keysym);
         }
         else
-            event_key_callback(ev, &globalconf.keys, 0, 0, &keysym);
+            event_key_callback(ev, &globalconf.keys, L, 0, 0, &keysym);
     }
 }
 
@@ -661,6 +735,7 @@ static void
 event_handle_maprequest(xcb_map_request_event_t *ev)
 {
     client_t *c;
+    xembed_window_t *em;
     xcb_get_window_attributes_cookie_t wa_c;
     xcb_get_window_attributes_reply_t *wa_r;
     xcb_get_geometry_cookie_t geom_c;
@@ -674,19 +749,27 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
     if(wa_r->override_redirect)
         goto bailout;
 
-    if(xembed_getbywin(&globalconf.embedded, ev->window))
+    if((em = xembed_getbywin(&globalconf.embedded, ev->window)))
     {
         xcb_map_window(globalconf.connection, ev->window);
         xembed_window_activate(globalconf.connection, ev->window);
+        /* The correct way to set this is via the _XEMBED_INFO property. Neither
+         * of the XEMBED not the systray spec talk about mapping windows.
+         * Apparently, Qt doesn't care and does not set an _XEMBED_INFO
+         * property. Let's simulate the XEMBED_MAPPED bit.
+         */
+        em->info.flags |= XEMBED_MAPPED;
+        luaA_systray_invalidate();
     }
     else if((c = client_getbywin(ev->window)))
     {
         /* Check that it may be visible, but not asked to be hidden */
-        if(client_maybevisible(c) && !c->hidden)
+        if(client_on_selected_tags(c) && !c->hidden)
         {
-            luaA_object_push(globalconf.L, c);
-            client_set_minimized(globalconf.L, -1, false);
-            lua_pop(globalconf.L, 1);
+            lua_State *L = globalconf_get_lua_State();
+            luaA_object_push(L, c);
+            client_set_minimized(L, -1, false);
+            lua_pop(L, 1);
             /* it will be raised, so just update ourself */
             client_raise(c);
         }
@@ -700,7 +783,7 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
             goto bailout;
         }
 
-        client_manage(ev->window, geom_r, false);
+        client_manage(ev->window, geom_r, wa_r);
 
         p_delete(&geom_r);
     }
@@ -719,14 +802,6 @@ event_handle_unmapnotify(xcb_unmap_notify_event_t *ev)
 
     if((c = client_getbywin(ev->window)))
         client_unmanage(c, true);
-    else
-        for(int i = 0; i < globalconf.embedded.len; i++)
-            if(globalconf.embedded.tab[i].win == ev->window)
-            {
-                xembed_window_array_take(&globalconf.embedded, i);
-                xcb_change_save_set(globalconf.connection, XCB_SET_MODE_DELETE, ev->window);
-                luaA_systray_invalidate();
-            }
 }
 
 /** The randr screen change notify event handler.
@@ -735,26 +810,66 @@ event_handle_unmapnotify(xcb_unmap_notify_event_t *ev)
 static void
 event_handle_randr_screen_change_notify(xcb_randr_screen_change_notify_event_t *ev)
 {
-    /* Code  of  XRRUpdateConfiguration Xlib  function  ported to  XCB
-     * (only the code relevant  to RRScreenChangeNotify) as the latter
-     * doesn't provide this kind of function */
-    if(ev->rotation & (XCB_RANDR_ROTATION_ROTATE_90 | XCB_RANDR_ROTATION_ROTATE_270))
-        xcb_randr_set_screen_size(globalconf.connection, ev->root, ev->height, ev->width,
-                                  ev->mheight, ev->mwidth);
-    else
-        xcb_randr_set_screen_size(globalconf.connection, ev->root, ev->width, ev->height,
-                                  ev->mwidth, ev->mheight);
+    /* Ignore events for other roots (do we get them at all?) */
+    if (ev->root != globalconf.screen->root)
+        return;
 
-    /* XRRUpdateConfiguration also executes the following instruction
-     * but it's not useful because SubpixelOrder is not used at all at
-     * the moment
-     *
-     * XRenderSetSubpixelOrder(dpy, snum, scevent->subpixel_order);
-     */
+    /* Do (part of) what XRRUpdateConfiguration() would do (update our state) */
+    if (ev->rotation & (XCB_RANDR_ROTATION_ROTATE_90 | XCB_RANDR_ROTATION_ROTATE_270)) {
+        globalconf.screen->width_in_pixels = ev->height;
+        globalconf.screen->height_in_pixels = ev->width;
+    } else {
+        globalconf.screen->width_in_pixels = ev->width;
+        globalconf.screen->height_in_pixels = ev->height;
+    }
 
-    awesome_restart();
+    globalconf.screen_need_refresh = true;
 }
 
+/** XRandR event handler for RRNotify subtype XRROutputChangeNotifyEvent
+ */
+static void
+event_handle_randr_output_change_notify(xcb_randr_notify_event_t *ev)
+{
+    if(ev->subCode == XCB_RANDR_NOTIFY_OUTPUT_CHANGE) {
+        xcb_randr_output_t output = ev->u.oc.output;
+        uint8_t connection = ev->u.oc.connection;
+        const char *connection_str = NULL;
+        xcb_randr_get_output_info_reply_t *info = NULL;
+        lua_State *L = globalconf_get_lua_State();
+
+        info = xcb_randr_get_output_info_reply(globalconf.connection,
+            xcb_randr_get_output_info_unchecked(globalconf.connection,
+                output,
+                XCB_CURRENT_TIME),
+            NULL);
+        if(!info)
+            return;
+
+        switch(connection) {
+            case XCB_RANDR_CONNECTION_CONNECTED:
+                connection_str = "Connected";
+                break;
+            case XCB_RANDR_CONNECTION_DISCONNECTED:
+                connection_str = "Disconnected";
+                break;
+            default:
+                connection_str = "Unknown";
+                break;
+        }
+
+        lua_pushlstring(L, (char *)xcb_randr_get_output_info_name(info), xcb_randr_get_output_info_name_length(info));
+        lua_pushstring(L, connection_str);
+        signal_object_emit(L, &global_signals, "screen::change", 2);
+
+        p_delete(&info);
+
+        /* The docs for RRSetOutputPrimary say we get this signal */
+        screen_update_primary();
+    }
+}
+
+
 /** The shape notify event handler.
  * \param ev The event.
  */
@@ -764,12 +879,13 @@ event_handle_shape_notify(xcb_shape_notify_event_t *ev)
     client_t *c = client_getbywin(ev->affected_window);
     if (c)
     {
-        luaA_object_push(globalconf.L, c);
+        lua_State *L = globalconf_get_lua_State();
+        luaA_object_push(L, c);
         if (ev->shape_kind == XCB_SHAPE_SK_BOUNDING)
-            luaA_object_emit_signal(globalconf.L, -1, "property::shape_client_bounding", 0);
+            luaA_object_emit_signal(L, -1, "property::shape_client_bounding", 0);
         if (ev->shape_kind == XCB_SHAPE_SK_CLIP)
-            luaA_object_emit_signal(globalconf.L, -1, "property::shape_client_clip", 0);
-        lua_pop(globalconf.L, 1);
+            luaA_object_emit_signal(L, -1, "property::shape_client_clip", 0);
+        lua_pop(L, 1);
     }
 }
 
@@ -790,9 +906,10 @@ event_handle_clientmessage(xcb_client_message_event_t *ev)
            && ev->format == 32
            && ev->data.data32[0] == XCB_ICCCM_WM_STATE_ICONIC)
         {
-            luaA_object_push(globalconf.L, c);
-            client_set_minimized(globalconf.L, -1, true);
-            lua_pop(globalconf.L, 1);
+            lua_State *L = globalconf_get_lua_State();
+            luaA_object_push(L, c);
+            client_set_minimized(L, -1, true);
+            lua_pop(L, 1);
         }
     }
     else if(ev->type == _XEMBED)
@@ -809,34 +926,8 @@ event_handle_clientmessage(xcb_client_message_event_t *ev)
 static void
 event_handle_mappingnotify(xcb_mapping_notify_event_t *ev)
 {
-    if(ev->request == XCB_MAPPING_MODIFIER
-       || ev->request == XCB_MAPPING_KEYBOARD)
-    {
-        xcb_get_modifier_mapping_cookie_t xmapping_cookie =
-            xcb_get_modifier_mapping_unchecked(globalconf.connection);
-
-        /* Free and then allocate the key symbols */
-        xcb_key_symbols_free(globalconf.keysyms);
-        globalconf.keysyms = xcb_key_symbols_alloc(globalconf.connection);
-
-        xutil_lock_mask_get(globalconf.connection, xmapping_cookie,
-                            globalconf.keysyms, &globalconf.numlockmask,
-                            &globalconf.shiftlockmask, &globalconf.capslockmask,
-                            &globalconf.modeswitchmask);
-
-        /* regrab everything */
-        xcb_screen_t *s = globalconf.screen;
-        /* yes XCB_BUTTON_MASK_ANY is also for grab_key even if it's look weird */
-        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
-        xwindow_grabkeys(s->root, &globalconf.keys);
-
-        foreach(_c, globalconf.clients)
-        {
-            client_t *c = *_c;
-            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->frame_window, XCB_BUTTON_MASK_ANY);
-            xwindow_grabkeys(c->frame_window, &c->keys);
-        }
-    }
+    /* Since we use XKB, we shouldn't get this event */
+    warn("Unexpected MappingNotify of type %d", ev->request);
 }
 
 static void
@@ -851,6 +942,26 @@ event_handle_reparentnotify(xcb_reparent_notify_event_t *ev)
         if (ev->parent != globalconf.screen->root)
             client_unmanage(c, true);
     }
+    else if (ev->parent != globalconf.systray.window) {
+        /* Embedded window moved elsewhere, end of embedding */
+        for(int i = 0; i < globalconf.embedded.len; i++)
+            if(globalconf.embedded.tab[i].win == ev->window)
+            {
+                xembed_window_array_take(&globalconf.embedded, i);
+                xcb_change_save_set(globalconf.connection, XCB_SET_MODE_DELETE, ev->window);
+                luaA_systray_invalidate();
+            }
+    }
+}
+
+static void
+event_handle_selectionclear(xcb_selection_clear_event_t *ev)
+{
+    if(ev->selection == globalconf.selection_atom)
+    {
+        warn("Lost WM_Sn selection, exiting...");
+        g_main_loop_quit(globalconf.loop);
+    }
 }
 
 /** \brief awesome xerror function.
@@ -880,10 +991,43 @@ xerror(xcb_generic_error_t *e)
     return;
 }
 
+static bool
+should_ignore(xcb_generic_event_t *event)
+{
+    uint8_t response_type = XCB_EVENT_RESPONSE_TYPE(event);
+
+    /* Remove completed sequences */
+    uint32_t sequence = event->full_sequence;
+    while (globalconf.ignore_enter_leave_events.len > 0) {
+        uint32_t end = globalconf.ignore_enter_leave_events.tab[0].end.sequence;
+        /* Do if (end >= sequence) break;, but handle wrap-around: The above is
+         * equivalent to end-sequence > 0 (assuming unlimited precision). With
+         * int32_t, this would mean that the sign bit is cleared, which means:
+         */
+        if (end - sequence < UINT32_MAX / 2)
+            break;
+        sequence_pair_array_take(&globalconf.ignore_enter_leave_events, 0);
+    }
+
+    /* Check if this event should be ignored */
+    if ((response_type == XCB_ENTER_NOTIFY || response_type == XCB_LEAVE_NOTIFY)
+            && globalconf.ignore_enter_leave_events.len > 0) {
+        uint32_t begin = globalconf.ignore_enter_leave_events.tab[0].begin.sequence;
+        uint32_t end   = globalconf.ignore_enter_leave_events.tab[0].end.sequence;
+        if (sequence >= begin && sequence <= end)
+            return true;
+    }
+
+    return false;
+}
+
 void event_handle(xcb_generic_event_t *event)
 {
     uint8_t response_type = XCB_EVENT_RESPONSE_TYPE(event);
 
+    if (should_ignore(event))
+        return;
+
     if(response_type == 0)
     {
         /* This is an error, not a event */
@@ -912,33 +1056,36 @@ void event_handle(xcb_generic_event_t *event)
         EVENT(XCB_PROPERTY_NOTIFY, property_handle_propertynotify);
         EVENT(XCB_REPARENT_NOTIFY, event_handle_reparentnotify);
         EVENT(XCB_UNMAP_NOTIFY, event_handle_unmapnotify);
+        EVENT(XCB_SELECTION_CLEAR, event_handle_selectionclear);
 #undef EVENT
     }
 
-    static uint8_t randr_screen_change_notify = 0;
-    static uint8_t shape_notify = 0;
+#define EXTENSION_EVENT(base, offset, callback) \
+    if (globalconf.event_base_ ## base != 0 \
+            && response_type == globalconf.event_base_ ## base + (offset)) \
+        callback((void *) event)
+    EXTENSION_EVENT(randr, XCB_RANDR_SCREEN_CHANGE_NOTIFY, event_handle_randr_screen_change_notify);
+    EXTENSION_EVENT(randr, XCB_RANDR_NOTIFY, event_handle_randr_output_change_notify);
+    EXTENSION_EVENT(shape, XCB_SHAPE_NOTIFY, event_handle_shape_notify);
+    EXTENSION_EVENT(xkb, 0, event_handle_xkb_notify);
+#undef EXTENSION_EVENT
+}
 
-    if(randr_screen_change_notify == 0)
-    {
-        /* check for randr extension */
-        const xcb_query_extension_reply_t *randr_query;
-        randr_query = xcb_get_extension_data(globalconf.connection, &xcb_randr_id);
-        if(randr_query->present)
-            randr_screen_change_notify = randr_query->first_event + XCB_RANDR_SCREEN_CHANGE_NOTIFY;
-    }
-    if(shape_notify == 0)
-    {
-        /* check for shape extension */
-        const xcb_query_extension_reply_t *shape_query;
-        shape_query = xcb_get_extension_data(globalconf.connection, &xcb_shape_id);
-        if(shape_query->present)
-            shape_notify = shape_query->first_event + XCB_SHAPE_NOTIFY;
-    }
+void event_init(void)
+{
+    const xcb_query_extension_reply_t *reply;
 
-    if (response_type == randr_screen_change_notify)
-        event_handle_randr_screen_change_notify((void *) event);
-    if (response_type == shape_notify)
-        event_handle_shape_notify((void *) event);
+    reply = xcb_get_extension_data(globalconf.connection, &xcb_randr_id);
+    if (reply && reply->present)
+        globalconf.event_base_randr = reply->first_event;
+
+    reply = xcb_get_extension_data(globalconf.connection, &xcb_shape_id);
+    if (reply && reply->present)
+        globalconf.event_base_shape = reply->first_event;
+
+    reply = xcb_get_extension_data(globalconf.connection, &xcb_xkb_id);
+    if (reply && reply->present)
+        globalconf.event_base_xkb = reply->first_event;
 }
 
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
