@@ -20,20 +20,125 @@
  *
  */
 
-#include "screen.h"
+/** awesome drawin API
+ *
+ * Furthermore to the classes described here, one can also use signals as
+ * described in @{signals} and X properties as described in @{xproperties}.
+ *
+ * @author Julien Danjou &lt;julien@danjou.info&gt;
+ * @copyright 2008-2009 Julien Danjou
+ * @classmod drawin
+ */
+
 #include "drawin.h"
-#include "objects/client.h"
-#include "screen.h"
-#include "xwindow.h"
-#include "luaa.h"
-#include "ewmh.h"
-#include "systray.h"
+#include "common/atoms.h"
 #include "common/xcursor.h"
 #include "common/xutil.h"
+#include "event.h"
+#include "ewmh.h"
+#include "objects/client.h"
+#include "objects/screen.h"
+#include "systray.h"
+#include "xwindow.h"
+
+#include "math.h"
 
 #include <cairo-xcb.h>
 #include <xcb/shape.h>
 
+/** Drawin object.
+ *
+ * @field border_width Border width.
+ * @field border_color Border color.
+ * @field ontop On top of other windows.
+ * @field cursor The mouse cursor.
+ * @field visible Visibility.
+ * @field opacity The opacity of the drawin, between 0 and 1.
+ * @field type The window type (desktop, normal, dock, …).
+ * @field x The x coordinates.
+ * @field y The y coordinates.
+ * @field width The width of the drawin.
+ * @field height The height of the drawin.
+ * @field drawable The drawin's drawable.
+ * @field window The X window id.
+ * @field shape_bounding The drawin's bounding shape as a (native) cairo surface.
+ * @field shape_clip The drawin's clip shape as a (native) cairo surface.
+ * @table drawin
+ */
+
+/**
+ * @signal property::geometry
+ */
+
+/**
+ * @signal property::shape_bounding
+ */
+
+/**
+ * @signal property::shape_clip
+ */
+
+/**
+ * @signal property::border_width
+ */
+
+/**
+ * @signal property::cursor
+ */
+
+/**
+ * @signal property::height
+ */
+
+/**
+ * @signal property::ontop
+ */
+
+/**
+ * @signal property::visible
+ */
+
+/**
+ * @signal property::width
+ */
+
+/**
+ * @signal property::x
+ */
+
+/**
+ * @signal property::y
+ */
+
+/** Get or set mouse buttons bindings to a drawin.
+ *
+ * @param buttons_table A table of buttons objects, or nothing.
+ * @function buttons
+ */
+
+/** Get or set drawin struts.
+ *
+ * @param strut A table with new strut, or nothing
+ * @return The drawin strut in a table.
+ * @function struts
+ */
+
+/** Get the number of instances.
+ *
+ * @return The number of drawin objects alive.
+ * @function instances
+ */
+
+/** Set a __index metamethod for all drawin instances.
+ * @tparam function cb The meta-method
+ * @function set_index_miss_handler
+ */
+
+/** Set a __newindex metamethod for all drawin instances.
+ * @tparam function cb The meta-method
+ * @function set_newindex_miss_handler
+ */
+
 LUA_OBJECT_FUNCS(drawin_class, drawin_t, drawin)
 
 /** Kick out systray windows.
@@ -73,18 +178,13 @@ drawin_wipe(drawin_t *w)
     w->drawable = NULL;
 }
 
-void
-drawin_unref_simplified(drawin_t **item)
-{
-    luaA_object_unref(globalconf.L, *item);
-}
-
 static void
-drawin_update_drawing(drawin_t *w, int widx)
+drawin_update_drawing(lua_State *L, int widx)
 {
-    luaA_object_push_item(globalconf.L, widx, w->drawable);
-    drawable_set_geometry(w->drawable, -1, w->geometry);
-    lua_pop(globalconf.L, 1);
+    drawin_t *w = luaA_checkudata(L, widx, &drawin_class);
+    luaA_object_push_item(L, widx, w->drawable);
+    drawable_set_geometry(L, -1, w->geometry);
+    lua_pop(L, 1);
 }
 
 /** Refresh the window content by copying its pixmap data to its window.
@@ -96,6 +196,37 @@ drawin_refresh_pixmap(drawin_t *w)
     drawin_refresh_pixmap_partial(w, 0, 0, w->geometry.width, w->geometry.height);
 }
 
+static void
+drawin_apply_moveresize(drawin_t *w)
+{
+    if (!w->geometry_dirty)
+        return;
+
+    w->geometry_dirty = false;
+    client_ignore_enterleave_events();
+    xcb_configure_window(globalconf.connection, w->window, 
+                         XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y
+                         | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
+                         (const uint32_t [])
+                         {
+                             w->geometry.x,
+                             w->geometry.y,
+                             w->geometry.width,
+                             w->geometry.height
+                         });
+    client_restore_enterleave_events();
+}
+
+void
+drawin_refresh(void)
+{
+    foreach(item, globalconf.drawins)
+    {
+        drawin_apply_moveresize(*item);
+        window_border_refresh((window_t *) *item);
+    }
+}
+
 /** Move and/or resize a drawin
  * \param L The Lua VM state.
  * \param udx The index of the drawin.
@@ -105,52 +236,35 @@ static void
 drawin_moveresize(lua_State *L, int udx, area_t geometry)
 {
     drawin_t *w = luaA_checkudata(L, udx, &drawin_class);
-    int number_of_vals = 0;
-    uint32_t moveresize_win_vals[4], mask_vals = 0;
+    area_t old_geometry = w->geometry;
 
-    if(w->geometry.x != geometry.x)
-    {
-        w->geometry.x = moveresize_win_vals[number_of_vals++] = geometry.x;
-        mask_vals |= XCB_CONFIG_WINDOW_X;
-    }
+    w->geometry = geometry;
+    if(w->geometry.width <= 0)
+        w->geometry.width = old_geometry.width;
+    if(w->geometry.height <= 0)
+        w->geometry.height = old_geometry.height;
 
-    if(w->geometry.y != geometry.y)
-    {
-        w->geometry.y = moveresize_win_vals[number_of_vals++] = geometry.y;
-        mask_vals |= XCB_CONFIG_WINDOW_Y;
-    }
+    w->geometry_dirty = true;
+    drawin_update_drawing(L, udx);
 
-    if(geometry.width > 0 && w->geometry.width != geometry.width)
-    {
-        w->geometry.width = moveresize_win_vals[number_of_vals++] = geometry.width;
-        mask_vals |= XCB_CONFIG_WINDOW_WIDTH;
-    }
-
-    if(geometry.height > 0 && w->geometry.height != geometry.height)
-    {
-        w->geometry.height = moveresize_win_vals[number_of_vals++] = geometry.height;
-        mask_vals |= XCB_CONFIG_WINDOW_HEIGHT;
-    }
-
-    drawin_update_drawing(w, udx);
-
-    /* Activate BMA */
-    client_ignore_enterleave_events();
-
-    if(mask_vals)
-        xcb_configure_window(globalconf.connection, w->window, mask_vals, moveresize_win_vals);
-
-    /* Deactivate BMA */
-    client_restore_enterleave_events();
-
-    if(mask_vals & XCB_CONFIG_WINDOW_X)
+    if (!AREA_EQUAL(old_geometry, w->geometry))
+        luaA_object_emit_signal(L, udx, "property::geometry", 0);
+    if (old_geometry.x != w->geometry.x)
         luaA_object_emit_signal(L, udx, "property::x", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_Y)
+    if (old_geometry.y != w->geometry.y)
         luaA_object_emit_signal(L, udx, "property::y", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_WIDTH)
+    if (old_geometry.width != w->geometry.width)
         luaA_object_emit_signal(L, udx, "property::width", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_HEIGHT)
+    if (old_geometry.height != w->geometry.height)
         luaA_object_emit_signal(L, udx, "property::height", 0);
+
+    screen_t *old_screen = screen_getbycoord(old_geometry.x, old_geometry.y);
+    screen_t *new_screen = screen_getbycoord(w->geometry.x, w->geometry.y);
+    if (old_screen != new_screen && strut_has_value(&w->strut))
+    {
+        screen_update_workarea(old_screen);
+        screen_update_workarea(new_screen);
+    }
 }
 
 /** Refresh the window content by copying its pixmap data to its window.
@@ -168,6 +282,9 @@ drawin_refresh_pixmap_partial(drawin_t *drawin,
     if (!drawin->drawable || !drawin->drawable->pixmap || !drawin->drawable->refreshed)
         return;
 
+    /* Make sure it really has the size it should have */
+    drawin_apply_moveresize(drawin);
+
     /* Make cairo do all pending drawing */
     cairo_surface_flush(drawin->drawable->surface);
     xcb_copy_area(globalconf.connection, drawin->drawable->pixmap,
@@ -176,8 +293,11 @@ drawin_refresh_pixmap_partial(drawin_t *drawin,
 }
 
 static void
-drawin_map(drawin_t *drawin, int widx)
+drawin_map(lua_State *L, int widx)
 {
+    drawin_t *drawin = luaA_checkudata(L, widx, &drawin_class);
+    /* Apply any pending changes */
+    drawin_apply_moveresize(drawin);
     /* Activate BMA */
     client_ignore_enterleave_events();
     /* Map the drawin */
@@ -190,7 +310,7 @@ drawin_map(drawin_t *drawin, int widx)
     drawin_array_append(&globalconf.drawins, drawin);
     /* Make sure it has a surface */
     if(drawin->drawable->surface == NULL)
-        drawin_update_drawing(drawin, widx);
+        drawin_update_drawing(L, widx);
 }
 
 static void
@@ -233,11 +353,11 @@ drawin_set_visible(lua_State *L, int udx, bool v)
 
         if(drawin->visible)
         {
-            drawin_map(drawin, udx);
+            drawin_map(L, udx);
             /* duplicate drawin */
             lua_pushvalue(L, udx);
             /* ref it */
-            luaA_object_ref_class(globalconf.L, -1, &drawin_class);
+            luaA_object_ref_class(L, -1, &drawin_class);
         }
         else
         {
@@ -248,15 +368,14 @@ drawin_set_visible(lua_State *L, int udx, bool v)
             /* Active BMA */
             client_restore_enterleave_events();
             /* unref it */
-            luaA_object_unref(globalconf.L, drawin);
+            luaA_object_unref(L, drawin);
         }
 
         luaA_object_emit_signal(L, udx, "property::visible", 0);
         if(strut_has_value(&drawin->strut))
         {
-            screen_t *screen =
-                screen_getbycoord(drawin->geometry.x, drawin->geometry.y);
-            screen_emit_signal(globalconf.L, screen, "property::workarea", 0);
+            screen_update_workarea(
+                    screen_getbycoord(drawin->geometry.x, drawin->geometry.y));
         }
     }
 }
@@ -273,6 +392,7 @@ drawin_allocator(lua_State *L)
     w->cursor = a_strdup("left_ptr");
     w->geometry.width = 1;
     w->geometry.height = 1;
+    w->geometry_dirty = false;
     w->type = _NET_WM_WINDOW_TYPE_NORMAL;
 
     drawable_allocator(L, (drawable_refresh_callback *) drawin_refresh_pixmap, w);
@@ -294,11 +414,14 @@ drawin_allocator(lua_State *L)
                           XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT
                           | XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY | XCB_EVENT_MASK_ENTER_WINDOW
                           | XCB_EVENT_MASK_LEAVE_WINDOW | XCB_EVENT_MASK_STRUCTURE_NOTIFY
-                          | XCB_EVENT_MASK_POINTER_MOTION | XCB_EVENT_MASK_EXPOSURE
+                          | XCB_EVENT_MASK_POINTER_MOTION | XCB_EVENT_MASK_BUTTON_PRESS
+                          | XCB_EVENT_MASK_BUTTON_RELEASE | XCB_EVENT_MASK_EXPOSURE
                           | XCB_EVENT_MASK_PROPERTY_CHANGE,
                           globalconf.default_cmap,
                           xcursor_new(globalconf.cursor_ctx, xcursor_font_fromstr(w->cursor))
                       });
+    xwindow_set_class_instance(w->window);
+    xwindow_set_name_static(w->window, "Awesome drawin");
 
     /* Set the right properties */
     ewmh_update_window_type(w->window, window_translate_type(w->type));
@@ -319,12 +442,12 @@ luaA_drawin_new(lua_State *L)
     return 1;
 }
 
-/* Set or get the drawin geometry.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam An optional table with drawin geometry.
- * \lreturn The drawin geometry.
+/** Get or set drawin geometry. That's the same as accessing or setting the x,
+ * y, width or height properties of a drawin.
+ *
+ * @param A table with coordinates to modify.
+ * @return A table with drawin coordinates and geometry.
+ * @function geometry
  */
 static int
 luaA_drawin_geometry(lua_State *L)
@@ -336,10 +459,10 @@ luaA_drawin_geometry(lua_State *L)
         area_t wingeom;
 
         luaA_checktable(L, 2);
-        wingeom.x = luaA_getopt_number(L, 2, "x", drawin->geometry.x);
-        wingeom.y = luaA_getopt_number(L, 2, "y", drawin->geometry.y);
-        wingeom.width = luaA_getopt_number(L, 2, "width", drawin->geometry.width);
-        wingeom.height = luaA_getopt_number(L, 2, "height", drawin->geometry.height);
+        wingeom.x = round(luaA_getopt_number_range(L, 2, "x", drawin->geometry.x, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        wingeom.y = round(luaA_getopt_number_range(L, 2, "y", drawin->geometry.y, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        wingeom.width = ceil(luaA_getopt_number_range(L, 2, "width", drawin->geometry.width, MIN_X11_SIZE, MAX_X11_SIZE));
+        wingeom.height = ceil(luaA_getopt_number_range(L, 2, "height", drawin->geometry.height, MIN_X11_SIZE, MAX_X11_SIZE));
 
         if(wingeom.width > 0 && wingeom.height > 0)
             drawin_moveresize(L, 1, wingeom);
@@ -356,7 +479,8 @@ LUA_OBJECT_EXPORT_PROPERTY(drawin, drawin_t, visible, lua_pushboolean)
 static int
 luaA_drawin_set_x(lua_State *L, drawin_t *drawin)
 {
-    drawin_moveresize(L, -3, (area_t) { .x = luaL_checknumber(L, -1),
+    int x = round(luaA_checknumber_range(L, -1, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+    drawin_moveresize(L, -3, (area_t) { .x = x,
                                         .y = drawin->geometry.y,
                                         .width = drawin->geometry.width,
                                         .height = drawin->geometry.height });
@@ -366,15 +490,16 @@ luaA_drawin_set_x(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_x(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.x);
+    lua_pushinteger(L, drawin->geometry.x);
     return 1;
 }
 
 static int
 luaA_drawin_set_y(lua_State *L, drawin_t *drawin)
 {
+    int y = round(luaA_checknumber_range(L, -1, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
-                                        .y = luaL_checknumber(L, -1),
+                                        .y = y,
                                         .width = drawin->geometry.width,
                                         .height = drawin->geometry.height });
     return 0;
@@ -383,16 +508,14 @@ luaA_drawin_set_y(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_y(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.y);
+    lua_pushinteger(L, drawin->geometry.y);
     return 1;
 }
 
 static int
 luaA_drawin_set_width(lua_State *L, drawin_t *drawin)
 {
-    int width = luaL_checknumber(L, -1);
-    if(width <= 0)
-        luaL_error(L, "invalid width");
+    int width = ceil(luaA_checknumber_range(L, -1, MIN_X11_SIZE, MAX_X11_SIZE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
                                         .y = drawin->geometry.y,
                                         .width = width,
@@ -403,16 +526,14 @@ luaA_drawin_set_width(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_width(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.width);
+    lua_pushinteger(L, drawin->geometry.width);
     return 1;
 }
 
 static int
 luaA_drawin_set_height(lua_State *L, drawin_t *drawin)
 {
-    int height = luaL_checknumber(L, -1);
-    if(height <= 0)
-        luaL_error(L, "invalid height");
+    int height = ceil(luaA_checknumber_range(L, -1, MIN_X11_SIZE, MAX_X11_SIZE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
                                        .y = drawin->geometry.y,
                                        .width = drawin->geometry.width,
@@ -423,7 +544,7 @@ luaA_drawin_set_height(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_height(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.height);
+    lua_pushinteger(L, drawin->geometry.height);
     return 1;
 }
 
@@ -520,6 +641,10 @@ luaA_drawin_set_shape_bounding(lua_State *L, drawin_t *drawin)
     cairo_surface_t *surf = NULL;
     if(!lua_isnil(L, -1))
         surf = (cairo_surface_t *)lua_touserdata(L, -1);
+
+    /* The drawin might have been resized to a larger size. Apply that. */
+    drawin_apply_moveresize(drawin);
+
     xwindow_set_shape(drawin->window,
             drawin->geometry.width + 2*drawin->border_width,
             drawin->geometry.height + 2*drawin->border_width,
@@ -555,6 +680,10 @@ luaA_drawin_set_shape_clip(lua_State *L, drawin_t *drawin)
     cairo_surface_t *surf = NULL;
     if(!lua_isnil(L, -1))
         surf = (cairo_surface_t *)lua_touserdata(L, -1);
+
+    /* The drawin might have been resized to a larger size. Apply that. */
+    drawin_apply_moveresize(drawin);
+
     xwindow_set_shape(drawin->window, drawin->geometry.width, drawin->geometry.height,
             XCB_SHAPE_SK_CLIP, surf, 0);
     luaA_object_emit_signal(L, -3, "property::shape_clip", 0);
@@ -629,17 +758,6 @@ drawin_class_setup(lua_State *L)
                             (lua_class_propfunc_t) luaA_drawin_set_shape_clip,
                             (lua_class_propfunc_t) luaA_drawin_get_shape_clip,
                             (lua_class_propfunc_t) luaA_drawin_set_shape_clip);
-
-    signal_add(&drawin_class.signals, "property::shape_bounding");
-    signal_add(&drawin_class.signals, "property::shape_clip");
-    signal_add(&drawin_class.signals, "property::border_width");
-    signal_add(&drawin_class.signals, "property::cursor");
-    signal_add(&drawin_class.signals, "property::height");
-    signal_add(&drawin_class.signals, "property::ontop");
-    signal_add(&drawin_class.signals, "property::visible");
-    signal_add(&drawin_class.signals, "property::width");
-    signal_add(&drawin_class.signals, "property::x");
-    signal_add(&drawin_class.signals, "property::y");
 }
 
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
