@@ -24,9 +24,7 @@
 #define AWESOME_OBJECTS_DRAWIN_H
 
 #include "objects/window.h"
-#include "common/luaobject.h"
 #include "objects/drawable.h"
-#include "draw.h"
 
 /** Drawin type */
 struct drawin_t
@@ -42,11 +40,11 @@ struct drawin_t
     drawable_t *drawable;
     /** The window geometry. */
     area_t geometry;
+    /** Do we have a pending geometry change that still needs to be applied? */
+    bool geometry_dirty;
 };
 
-void drawin_unref_simplified(drawin_t **);
-
-ARRAY_FUNCS(drawin_t *, drawin, drawin_unref_simplified)
+ARRAY_FUNCS(drawin_t *, drawin, DO_NOTHING)
 
 drawin_t * drawin_getbywin(xcb_window_t);
 
