@@ -19,18 +19,16 @@
  *
  */
 
-#include <xcb/xcb.h>
-#include <xcb/xcb_icccm.h>
-#include <xcb/xcb_atom.h>
-
-#include "luaa.h"
-#include "screen.h"
 #include "systray.h"
-#include "xwindow.h"
-#include "common/array.h"
 #include "common/atoms.h"
 #include "common/xutil.h"
 #include "objects/drawin.h"
+#include "xwindow.h"
+#include "globalconf.h"
+
+#include <xcb/xcb.h>
+#include <xcb/xcb_icccm.h>
+#include <xcb/xcb_atom.h>
 
 #define SYSTEM_TRAY_REQUEST_DOCK 0 /* Begin icon docking */
 
@@ -45,12 +43,16 @@ systray_init(void)
     xcb_screen_t *xscreen = globalconf.screen;
 
     globalconf.systray.window = xcb_generate_id(globalconf.connection);
+    globalconf.systray.background_pixel = xscreen->black_pixel;
     xcb_create_window(globalconf.connection, xscreen->root_depth,
                       globalconf.systray.window,
                       xscreen->root,
                       -1, -1, 1, 1, 0,
                       XCB_COPY_FROM_PARENT, xscreen->root_visual,
-                      0, NULL);
+                      XCB_CW_BACK_PIXEL | XCB_CW_EVENT_MASK, (const uint32_t [])
+                      { xscreen->black_pixel, XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT });
+    xwindow_set_class_instance(globalconf.systray.window);
+    xwindow_set_name_static(globalconf.systray.window, "Awesome systray window");
 
     atom_name = xcb_atom_name_by_screen("_NET_SYSTEM_TRAY", globalconf.default_screen);
     if(!atom_name)
@@ -122,11 +124,10 @@ systray_cleanup(void)
 
 /** Handle a systray request.
  * \param embed_win The window to embed.
- * \param info The embedding info
  * \return 0 on no error.
  */
 int
-systray_request_handle(xcb_window_t embed_win, xembed_info_t *info)
+systray_request_handle(xcb_window_t embed_win)
 {
     xembed_window_t em;
     xcb_get_property_cookie_t em_cookie;
@@ -143,8 +144,7 @@ systray_request_handle(xcb_window_t embed_win, xembed_info_t *info)
 
     p_clear(&em_cookie, 1);
 
-    if(!info)
-        em_cookie = xembed_info_get_unchecked(globalconf.connection, embed_win);
+    em_cookie = xembed_info_get_unchecked(globalconf.connection, embed_win);
 
     xcb_change_window_attributes(globalconf.connection, embed_win, XCB_CW_EVENT_MASK,
                                  select_input_val);
@@ -159,10 +159,11 @@ systray_request_handle(xcb_window_t embed_win, xembed_info_t *info)
 
     em.win = embed_win;
 
-    if(info)
-        em.info = *info;
-    else
-        xembed_info_get_reply(globalconf.connection, em_cookie, &em.info);
+    if (!xembed_info_get_reply(globalconf.connection, em_cookie, &em.info)) {
+        /* Set some sane defaults */
+        em.info.version = XEMBED_VERSION;
+        em.info.flags = XEMBED_MAPPED;
+    }
 
     xembed_embedded_notify(globalconf.connection, em.win,
                            globalconf.systray.window,
@@ -194,7 +195,7 @@ systray_process_client_message(xcb_client_message_event_t *ev)
             return -1;
 
         if(globalconf.screen->root == geom_r->root)
-            ret = systray_request_handle(ev->data.data32[2], NULL);
+            ret = systray_request_handle(ev->data.data32[2]);
 
         p_delete(&geom_r);
         break;
@@ -246,26 +247,42 @@ xembed_process_client_message(xcb_client_message_event_t *ev)
     return 0;
 }
 
+static int
+systray_num_visible_entries(void)
+{
+    int result = 0;
+    foreach(em, globalconf.embedded)
+        if (em->info.flags & XEMBED_MAPPED)
+            result++;
+    return result;
+}
+
 /** Inform lua that the systray needs to be updated.
  */
 void
 luaA_systray_invalidate(void)
 {
-    signal_object_emit(globalconf.L, &global_signals, "systray::update", 0);
+    lua_State *L = globalconf_get_lua_State();
+    signal_object_emit(L, &global_signals, "systray::update", 0);
+
+    /* Unmap now if the systray became empty */
+    if(systray_num_visible_entries() == 0)
+        xcb_unmap_window(globalconf.connection, globalconf.systray.window);
 }
 
 static void
-systray_update(int base_size, bool horizontal, bool reverse)
+systray_update(int base_size, bool horizontal, bool reverse, int spacing, bool force_redraw)
 {
     if(base_size <= 0)
         return;
 
     /* Give the systray window the correct size */
+    int num_entries = systray_num_visible_entries();
     uint32_t config_vals[4] = { base_size, base_size, 0, 0 };
     if(horizontal)
-        config_vals[0] = base_size * globalconf.embedded.len;
+        config_vals[0] = base_size * num_entries + spacing * (num_entries - 1);
     else
-        config_vals[1] = base_size * globalconf.embedded.len;
+        config_vals[1] = base_size * num_entries + spacing * (num_entries - 1);
     xcb_configure_window(globalconf.connection,
                          globalconf.systray.window,
                          XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
@@ -283,14 +300,22 @@ systray_update(int base_size, bool horizontal, bool reverse)
         else
             em = &globalconf.embedded.tab[i];
 
+        if (!(em->info.flags & XEMBED_MAPPED))
+        {
+            xcb_unmap_window(globalconf.connection, em->win);
+            continue;
+        }
+
         xcb_configure_window(globalconf.connection, em->win,
                              XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
                              config_vals);
         xcb_map_window(globalconf.connection, em->win);
+        if (force_redraw)
+            xcb_clear_area(globalconf.connection, 1, em->win, 0, 0, 0, 0);
         if(horizontal)
-            config_vals[0] += base_size;
+            config_vals[0] += base_size + spacing;
         else
-            config_vals[1] += base_size;
+            config_vals[1] += base_size + spacing;
     }
 }
 
@@ -302,9 +327,10 @@ systray_update(int base_size, bool horizontal, bool reverse)
  * \lparam x X position for the systray.
  * \lparam y Y position for the systray.
  * \lparam base_size The size (width and height) each systray item gets.
- * \lparam horiz If true, the systray is horizontal, else vertical
- * \lparam bg Color of the systray background
- * \lparam revers If true, the systray icon order will be reversed, else default
+ * \lparam horiz If true, the systray is horizontal, else vertical.
+ * \lparam bg Color of the systray background.
+ * \lparam revers If true, the systray icon order will be reversed, else default.
+ * \lparam spacing The size of the spacing between icons.
  */
 int
 luaA_systray(lua_State *L)
@@ -315,20 +341,26 @@ luaA_systray(lua_State *L)
     {
         size_t bg_len;
         drawin_t *w = luaA_checkudata(L, 1, &drawin_class);
-        int x = luaL_checknumber(L, 2);
-        int y = luaL_checknumber(L, 3);
-        int base_size = luaL_checknumber(L, 4);
+        int x = round(luaA_checknumber_range(L, 2, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        int y = round(luaA_checknumber_range(L, 3, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        int base_size = ceil(luaA_checknumber_range(L, 4, MIN_X11_SIZE, MAX_X11_SIZE));
         bool horiz = lua_toboolean(L, 5);
         const char *bg = luaL_checklstring(L, 6, &bg_len);
         bool revers = lua_toboolean(L, 7);
+        int spacing = ceil(luaA_checknumber_range(L, 8, 0, MAX_X11_COORDINATE));
         color_t bg_color;
+        bool force_redraw = false;
 
-        if(color_init_reply(color_init_unchecked(&bg_color, bg, bg_len)))
+        if(color_init_reply(color_init_unchecked(&bg_color, bg, bg_len))
+                && globalconf.systray.background_pixel != bg_color.pixel)
         {
             uint32_t config_back[] = { bg_color.pixel };
+            globalconf.systray.background_pixel = bg_color.pixel;
             xcb_change_window_attributes(globalconf.connection,
                                          globalconf.systray.window,
                                          XCB_CW_BACK_PIXEL, config_back);
+            xcb_clear_area(globalconf.connection, 1, globalconf.systray.window, 0, 0, 0, 0);
+            force_redraw = true;
         }
 
         if(globalconf.systray.parent != w)
@@ -347,18 +379,15 @@ luaA_systray(lua_State *L)
 
         globalconf.systray.parent = w;
 
-        if(globalconf.embedded.len != 0)
+        if(systray_num_visible_entries() != 0)
         {
-            systray_update(base_size, horiz, revers);
+            systray_update(base_size, horiz, revers, spacing, force_redraw);
             xcb_map_window(globalconf.connection,
                            globalconf.systray.window);
         }
-        else
-            xcb_unmap_window(globalconf.connection,
-                             globalconf.systray.window);
     }
 
-    lua_pushnumber(L, globalconf.embedded.len);
+    lua_pushinteger(L, systray_num_visible_entries());
     luaA_object_push(L, globalconf.systray.parent);
     return 2;
 }
