@@ -19,15 +19,14 @@
  *
  */
 
+#include "xwindow.h"
+#include "common/atoms.h"
+#include "objects/button.h"
+
 #include <xcb/xcb.h>
-#include <xcb/xcb_atom.h>
 #include <xcb/shape.h>
 #include <cairo-xcb.h>
 
-#include "xwindow.h"
-#include "objects/button.h"
-#include "common/atoms.h"
-
 /** Mask shorthands */
 #define BUTTONMASK     (XCB_EVENT_MASK_BUTTON_PRESS | XCB_EVENT_MASK_BUTTON_RELEASE)
 
@@ -96,7 +95,8 @@ xwindow_configure(xcb_window_t win, area_t geometry, int border)
     ce.border_width = border;
     ce.above_sibling = XCB_NONE;
     ce.override_redirect = false;
-    xcb_send_event(globalconf.connection, false, win, XCB_EVENT_MASK_STRUCTURE_NOTIFY, (char *) &ce);
+    xcb_send_event(globalconf.connection, false, win,
+                   XCB_EVENT_MASK_STRUCTURE_NOTIFY, (char *) &ce);
 }
 
 /** Grab or ungrab buttons on a window.
@@ -155,7 +155,8 @@ xwindow_grabkeys(xcb_window_t win, key_array_t *keys)
  * \param win The window
  * \return A cookie for xwindow_get_opacity_from_reply().
  */
-xcb_get_property_cookie_t xwindow_get_opacity_unchecked(xcb_window_t win)
+xcb_get_property_cookie_t
+xwindow_get_opacity_unchecked(xcb_window_t win)
 {
     return xcb_get_property_unchecked(globalconf.connection, false, win,
                                       _NET_WM_WINDOW_OPACITY, XCB_ATOM_CARDINAL, 0L, 1L);
@@ -330,6 +331,9 @@ xwindow_shape_pixmap(int width, int height, cairo_surface_t *surf)
     cairo_surface_t *dest;
     cairo_t *cr;
 
+    if (width <= 0 || height <= 0)
+        return XCB_NONE;
+
     xcb_create_pixmap(globalconf.connection, 1, pixmap, globalconf.screen->root, width, height);
     dest = cairo_xcb_surface_create_for_bitmap(globalconf.connection, globalconf.screen, pixmap, width, height);
 
